// g++ -D QUEUE_MAIN -std=c++0x Queue.cpp
// g++ -D QUEUE_MAIN -std=c++11 Queue.cpp


#include "Queue.hpp"

#include <cassert>
#include <cstddef>

#include <iostream>

template <typename T>
Queue<T>::Queue( void ) :
    N (0),
    first (nullptr),
    last (nullptr)
    {
    }
    
template <typename T>
Queue<T>::~Queue( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    while ( first ) {
        Node *next = first->next;
        delete first;
        first = next;
        }
    }

template <typename T>
void Queue<T>::enqueue( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = nullptr;

    if ( last ) {
        last->next = node;
        last = node;
        }
    else {
        last = first = node;
        }

    N++;
    }

template <typename T>
T Queue<T>::dequeue( void ) {
    assert( N > 0 );

    T item = first->item;

    Node *old = first;
    first = first->next;
    delete old;
    N--;

    if ( ! first )
        last = nullptr;

    return item;
    }

template <typename T>
bool Queue<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Queue<T>::size( void ) {
    return N;
    }

template <typename T>
typename Queue<T>::iterator Queue<T>::begin( void ) {
    return Queue<T>::iterator( first );
    }

template <typename T>
typename Queue<T>::iterator Queue<T>::end( void ) {
    return Queue<T>::iterator( nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Queue<T>::iterator::iterator( Node *c ) :
    curr (c)
    {
    }

template <typename T>
typename Queue<T>::iterator & Queue<T>::iterator::operator++() {
    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return *this;
    }

template <typename T>
typename Queue<T>::iterator Queue<T>::iterator::operator++( int ) {
    auto t = Queue<T>::iterator( *this );

    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return t;
    }

template <typename T>
T Queue<T>::iterator::operator*() {
    return this->curr->item;
    }

template <typename T>
bool Queue<T>::iterator::operator!=( typename Queue<T>::iterator other ) {
    return this->curr != other.curr;
    }


////////////////////////////////////////////////////////////////////////////////

#ifdef QUEUE_MAIN

#include <iostream>

int main ( int argc, char **argv ) {

    Queue<long> queue;

    long i;
    while ( ! std::cin.eof() ) {
        std::cin >> i;
        if ( std::cin.good() ) 
            if ( i >= 0 )
                queue.enqueue(i);
            else
                queue.dequeue();
        }

    std::cout << "Queue has " << queue.size() << " entries." << std::endl;

    for ( auto iter = queue.begin(); iter != queue.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

    std::cout << "Dequeuing entries..." << std::endl;

    while ( ! queue.is_empty() ) {
        i = queue.dequeue();
        std::cout << i << std::endl;
        }

    std::cout << "Queue has " << queue.size() << " entries." << std::endl;

    }

#endif
